from argparse import ArgumentParser
from pathlib import Path
from typing import Any, Sequence

from .config import CONFIG


CMD = 'command'
TRAIN = 'train'
DATA_DIR = 'data_dir'
SAVE_DIR = 'save_dir'
FILE_EXTENSIONS = 'file_extensions'
BATCH_SIZE = 'batch_size'
NUM_EPOCHS = 'num_epochs'
EARLY_STOPPING_PATIENCE = 'early_stopping_patience'

INFER = 'infer'
MODEL_DIR = 'model_dir'


def ext_list(string: str) -> list[str]:
    out = []
    for ext in string.split(','):
        ext = ext.strip()
        if not ext.startswith('.'):
            raise ValueError("Extensions must start with a dot")
        out.append(ext)
    return out


def parse_cli(args: Sequence[str] = None) -> dict[str, Any]:
    parser = ArgumentParser(
        prog=CONFIG.PROGRAM_NAME,
        description="Character CAPTCHA Solver",
    )
    parser.add_argument(
        '-E', f'--{FILE_EXTENSIONS.replace("_", "-")}',
        default=CONFIG.DEFAULT_DATA_FILE_EXTENSIONS,
        type=ext_list,
        help=f"When used in `{TRAIN}` mode, extensions of the image files to be used for training/testing the model. "
             f"When used in `{INFER}` mode, extensions of the image files to use the model on."
             f"Defaults to {CONFIG.DEFAULT_DATA_FILE_EXTENSIONS}."
    )
    subparsers = parser.add_subparsers(dest=CMD)

    parser_train = subparsers.add_parser(TRAIN, help="trains a new model")
    parser_train.add_argument(
        DATA_DIR,
        type=Path,
        help="Directory containing the image files to be used for training/testing the model."
    )
    parser_train.add_argument(
        '-s', f'--{SAVE_DIR.replace("_", "-")}',
        default=CONFIG.DEFAULT_SAVE_DIR,
        type=Path,
        help=f"Directory in which to save trained models. A subdirectory for each training session named with the "
             f"current date and time will be created there and the model will be saved in that subdirectory. "
             f"Defaults to '{CONFIG.DEFAULT_SAVE_DIR}'."
    )
    parser_train.add_argument(
        '-b', f'--{BATCH_SIZE.replace("_", "-")}',
        default=CONFIG.DEFAULT_BATCH_SIZE,
        type=int,
        help=f"The dataset will be divided into batches; this determines the number of images in each batch. "
             f"Defaults to {CONFIG.DEFAULT_BATCH_SIZE}."
    )
    parser_train.add_argument(
        '-n', f'--{NUM_EPOCHS.replace("_", "-")}',
        default=CONFIG.DEFAULT_NUM_EPOCHS,
        type=int,
        help=f"The number of training epochs. Defaults to {CONFIG.DEFAULT_NUM_EPOCHS}."
    )
    parser_train.add_argument(
        '-p', f'--{EARLY_STOPPING_PATIENCE.replace("_", "-")}',
        default=CONFIG.DEFAULT_EARLY_STOPPING_PATIENCE,
        type=int,
        help=f"The number of training epochs with no improvement over a previously achieved optimum to allow before "
             f"stopping training early (i.e. without completing all epochs). "
             f"Defaults to {CONFIG.DEFAULT_EARLY_STOPPING_PATIENCE}."
    )

    parser_infer = subparsers.add_parser(INFER, help="uses an existing model to make inferences")
    parser_infer.add_argument(
        MODEL_DIR,
        type=Path,
        help="Directory containing the model to use for inference."
    )
    parser_infer.add_argument(
        DATA_DIR,
        type=Path,
        help="Directory containing the image files to use the model on."
    )
    return vars(parser.parse_args(args))


def main() -> None:
    kwargs = parse_cli()
    cmd = kwargs.pop(CMD)
    if cmd == TRAIN:
        from .model import start
        start(**kwargs)
    elif cmd == INFER:
        from .infer import start
        start(**kwargs)
    else:
        raise NotImplemented


if __name__ == '__main__':
    main()
