from pathlib import Path
from typing import Iterable

import numpy as np
import tensorflow as tf
from keras.api._v2.keras.models import Model, load_model
from keras.api._v2.keras.layers.experimental.preprocessing import StringLookup
from keras.api._v2.keras.backend import ctc_decode

from .config import CONFIG
from .preproc import encode_image, decode_label, plot_images
from .types import PathT


def images_to_input(*images) -> tf.data.Dataset:
    array = np.array([encode_image(img) for img in images])
    return tf.data.Dataset.from_tensor_slices(array)


def decode_batch_outputs(predictions, backward_lookup_table: StringLookup) -> list[str]:
    input_len = np.ones(predictions.shape[0]) * predictions.shape[1]
    # Use greedy search. For complex tasks, you can use beam search
    sequences, _ = ctc_decode(predictions, input_length=input_len, greedy=True)
    results = sequences[0][:, :CONFIG.MAX_STRING_LENGTH]
    # Iterate over the results and get back the text
    return [decode_label(result, backward_lookup_table) for result in results]


def load_inference_model(path: PathT) -> Model:
    with open(Path(path, CONFIG.VOCABULARY_FILE_NAME), 'r') as vocab_file:
        backward_lookup_table = StringLookup(vocabulary=list(vocab_file.read()), mask_token=None, invert=True)
    saved_model = load_model(path)
    inference_model = Model(
        saved_model.get_layer(name=CONFIG.INPUT_LAYER_NAME_IMAGE).input,
        saved_model.get_layer(name=CONFIG.OUTPUT_LAYER_NAME).output
    )

    def infer_and_decode(x: tf.data.Dataset) -> list[str]:
        return decode_batch_outputs(predictions=inference_model.predict(x), backward_lookup_table=backward_lookup_table)

    inference_model.infer_and_decode = infer_and_decode
    inference_model.backward_lookup_table = backward_lookup_table
    return inference_model


def start(model_dir: PathT, data_dir: PathT,
          file_extensions: Iterable[str] = CONFIG.DEFAULT_DATA_FILE_EXTENSIONS) -> None:
    data_dir = Path(data_dir)
    file_paths = []
    for ext in file_extensions:
        file_paths.extend(data_dir.glob(f'*{ext}'))
    file_paths.sort()
    count = len(file_paths)
    if count > 24:
        raise ValueError("Too many files")
    # images = []
    # for path in file_paths:
    #     with open(path, 'rb') as f:
    #         images.append(f.read())
    dataset = images_to_input(*file_paths)
    model = load_inference_model(model_dir)
    labels = model.infer_and_decode(dataset.batch(count))
    plot_images(list(dataset.as_numpy_iterator()), labels=labels)
