import os
from datetime import datetime
from pathlib import Path
from typing import Iterable

import numpy as np
import tensorflow as tf
from tensorflow import keras
from tensorflow.keras import layers

from .config import CONFIG
from .preproc import DatasetsInterface
from .types import PathT


THIS_DIR = os.path.dirname(os.path.realpath(__file__))
# Build paths relative to this file's directory like this: os_path.join(THIS_DIR, ...)


class CTCLayer(layers.Layer):
    def __init__(self, name: str = None):
        super().__init__(name=name)
        self.loss_fn = keras.backend.ctc_batch_cost

    def call(self, y_true: np.ndarray = None, y_pred: np.ndarray = None) -> np.ndarray:
        # Compute the training-time loss value and add it
        # to the layer using `self.add_loss()`.
        batch_len = tf.cast(tf.shape(y_true)[0], dtype='int64')
        input_length = tf.cast(tf.shape(y_pred)[1], dtype='int64')
        label_length = tf.cast(tf.shape(y_true)[1], dtype='int64')
        input_length = input_length * tf.ones(shape=(batch_len, 1), dtype='int64')
        label_length = label_length * tf.ones(shape=(batch_len, 1), dtype='int64')
        loss = self.loss_fn(y_true, y_pred, input_length, label_length)
        self.add_loss(loss)
        # At test time, just return the computed predictions
        return y_pred


# Factor by which the image is going to be downsampled
# by the convolutional blocks. We will be using two
# convolution blocks and each block will have
# a pooling layer which downsample the features by a factor of 2.
# Hence total downsampling factor would be 4.
downsample_factor = 4


def build_model(alphabet_size: int,
                img_width: int = CONFIG.IMG_WIDTH,
                img_height: int = CONFIG.IMG_HEIGHT,
                optimizer: keras.optimizers.Optimizer = keras.optimizers.Adam()) -> keras.models.Model:
    # Inputs to the model
    input_img = layers.Input(
        shape=(img_width, img_height, 1),
        dtype='float32',
        name=CONFIG.INPUT_LAYER_NAME_IMAGE
    )
    labels = layers.Input(
        shape=(None, ),
        dtype='float32',
        name=CONFIG.INPUT_LAYER_NAME_LABEL,
    )
    # First conv block
    x = layers.Conv2D(
        filters=32,
        kernel_size=(3, 3),
        activation='relu',
        kernel_initializer='he_normal',
        padding='same',
        name='conv1',
    )(input_img)
    x = layers.MaxPooling2D(
        pool_size=(2, 2),
        name='pool1'
    )(x)
    # Second conv block
    x = layers.Conv2D(
        filters=64,
        kernel_size=(3, 3),
        activation='relu',
        kernel_initializer='he_normal',
        padding='same',
        name='conv2',
    )(x)
    x = layers.MaxPooling2D(
        pool_size=(2, 2),
        name='pool2'
    )(x)
    # We have used two max. pooling layers with pool size and strides 2.
    # Hence, downsampled feature maps are 4x smaller. The number of
    # filters in the last layer is 64. Reshape accordingly before
    # passing the output to the RNN part of the model
    new_shape = (
        (img_width // 4),
        (img_height // 4) * 64
    )
    x = layers.Reshape(
        target_shape=new_shape,
        name='reshape'
    )(x)
    x = layers.Dense(
        units=64,
        activation='relu',
        name='dense1'
    )(x)
    x = layers.Dropout(rate=0.2)(x)
    # RNNs
    x = layers.Bidirectional(
        layers.LSTM(
            units=128,
            return_sequences=True,
            dropout=0.25,
        )
    )(x)
    x = layers.Bidirectional(
        layers.LSTM(
            units=64,
            return_sequences=True,
            dropout=0.25,
        )
    )(x)
    # Output layer
    x = layers.Dense(
        units=alphabet_size + 1,
        activation='softmax',
        name=CONFIG.OUTPUT_LAYER_NAME,
    )(x)
    # Add CTC layer for calculating CTC loss at each step
    output = CTCLayer(name='ctc_loss')(labels, x)
    # Define the model
    model = keras.models.Model(
        inputs=[input_img, labels],
        outputs=output,
        name='ocr_model_v1'
    )
    # Compile the model and return
    model.compile(optimizer=optimizer)
    return model


def train_model(model: keras.models.Model, train_dataset: tf.data.Dataset, valid_dataset: tf.data.Dataset,
                num_epochs: int = CONFIG.DEFAULT_NUM_EPOCHS,
                early_stopping_patience: int = CONFIG.DEFAULT_EARLY_STOPPING_PATIENCE) -> keras.callbacks.History:
    # Add early stopping
    early_stopping = keras.callbacks.EarlyStopping(
        monitor='val_loss',
        patience=early_stopping_patience,
        restore_best_weights=True,
    )
    # Train the model
    history = model.fit(
        x=train_dataset,
        validation_data=valid_dataset,
        epochs=num_epochs,
        callbacks=[early_stopping],
    )
    return history


def start(data_dir: PathT, save_dir: PathT, file_extensions: Iterable[str] = CONFIG.DEFAULT_DATA_FILE_EXTENSIONS,
          batch_size: int = CONFIG.DEFAULT_BATCH_SIZE, num_epochs: int = CONFIG.DEFAULT_NUM_EPOCHS,
          early_stopping_patience: int = CONFIG.DEFAULT_EARLY_STOPPING_PATIENCE) -> None:
    save_dir = Path(save_dir, datetime.now().strftime('%Y-%m-%d_%H-%M'))
    save_dir.mkdir(parents=True)
    print("\nConstructing datasets\n")
    data_interface = DatasetsInterface(batch_size=int(batch_size), data_dir=data_dir, extensions=file_extensions)
    data_interface.split_and_make_datasets()
    print("\nBuilding model\n")
    model = build_model(len(data_interface.characters))
    print("\nBeginning training\n")
    train_model(model, data_interface.training, data_interface.validation, num_epochs=num_epochs,
                early_stopping_patience=early_stopping_patience)
    print("\nSaving model\n")
    model.save(save_dir)
    print("\nSaving vocabulary\n")
    vocabulary = ''.join(data_interface.forward_lookup_table.get_vocabulary())
    with open(os.path.join(save_dir, CONFIG.VOCABULARY_FILE_NAME), 'w') as f:
        f.write(vocabulary)
    print("\nAll saved!\n")
